// ZEVA EVSE Interface
// Open Source version, released under MIT License (see Readme file)
// Last modified by Ian Hooper (ZEVA), August 2021

// Code for ATmega16M1 (also suitable for ATmega32M1, ATmega64m1)
// Fuses: 8Mhz+ external crystal, CKDIV8 off, brownout 4.2V

#define PCB_VERSION	12		// XX = X.X, compatibility with older PCBs to allow for different pin assignments / functionality
#define SEND_DIAGNOSTIC_VAC	0 // This is just used for testing, sends VAC as DC current to confirm measurement accuracy

#define AC_VOLTAGE	240 // Hardcoded according to the AC voltage of the country it's used in, for calculating power available
	// Note that this only applies for PCB versions 1.0 and 1.1, or if nothing is connected to the VAC terminal

#define CAN_BAUD_RATE		250 // Code knows how to do 125, 250, 500, 1000kbps
#define USE_29BIT_IDS		1	// Or 0 for 11-bit IDs

#define CAN_ID	45	// For the transmitted message containing available power from charging station (in Watts)

#include <inttypes.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdbool.h>
#include <avr/wdt.h>

// Function declarations
void CanTX(long packetID, unsigned char* data, unsigned char length);
void SetupPorts();

// Global variables
uint8_t data[8];
volatile int ticks = 0;
volatile int pilotTimer = 0;
long smoothPilotTimer = 0;
unsigned short wattsAvailable = 0;
char evmsPresent = false;
char evmsCommsTimeout = 0;
char error = false;

enum { IDLE, CHARGING };
char state = IDLE;

SIGNAL(INT0_vect) // Interrupt function called upon any change of INT0 pin
{
	// Read PILOT pin PD6, rising edge starts timer, falling edge ends timer and gets signal on-time
	if (PIND & (1<<PD6))
		TCNT1 = 0;
	else
		pilotTimer = TCNT1;
}

SIGNAL(INT1_vect)
{
	// Does nothing, but has to be here for when INT1 generates an interrupt (used to wake from sleep)
}

SIGNAL(TIMER1_OVF_vect) // Timer1 overflow, probably indicates some problem with the pilot PWM signal
{
	pilotTimer = 0;
}

SIGNAL(TIMER0_OVF_vect) // Called at 31372Hz, increment ticks used for main loop timing
{
	ticks++;
}

SIGNAL(CAN_INT_vect) // Interrupt function for when new CAN message is received
{
	int8_t length, savedCANPage;
	uint8_t data[8];
	
	savedCANPage = CANPAGE; // Saves current MOB
	CANPAGE = CANHPMOB & 0xF0; // Selects MOB with highest priority interrupt
	if (CANSTMOB & (1<<RXOK))
	{
		length = CANCDMOB & 0x0F; // Number of bytes to receive is bottom four bits of this reg
		for (int8_t i=0; i<length; i++) data[i] = CANMSG; // This autoincrements when read

		// ID has top 8 bits in IDT1 and bottom 3 bits at TOP of IDT2
		long packetID;
		if (USE_29BIT_IDS)
			packetID = (((long)CANIDT1)<<21) + (((long)CANIDT2)<<13) + (CANIDT3<<5) + (CANIDT4>>3);
		else
			packetID = (CANIDT1<<3) + (CANIDT2>>5);

		// Check packets in here

		if (packetID == 30) // EVMS status - don't care what's in it, only whether it's there
		{
			evmsPresent = true;
			evmsCommsTimeout = 4;
		}
		
		CANCDMOB = (1<<CONMOB1) | (8<<DLC0) | ((1<<IDE) * USE_29BIT_IDS); // Enable reception, 11-bit IDE, data length 8
		// Note: The DLC field of CANCDMOB register is updated by the received MOB, and if it differs from above, an error is set
	}
	CANSTMOB = 0x00; // Reset interrupt reason on selected channel
	CANPAGE = savedCANPage;
}

#define ADC_VREF_TYPE (1<<REFS0) // AVCC ref with cap on AREF pin
int ReadADC(unsigned char channel)
{
	ADMUX = channel | ADC_VREF_TYPE;
	_delay_us(10);				// Delay needed for the stabilization of the ADC input voltage
	ADCSRA |= (1<<ADSC);				// Start the AD conversion
	while ((ADCSRA & 0x10) == 0);	// Wait for the AD conversion to complete
	ADCSRA |= 0x10;
	return ADCW;
}

int main()
{
	SetupPorts();
	for (int n=0; n<8; n++) data[n] = 0; // Initialise TX buffer to zero
	sei();

	unsigned short sleepTimer = 8; // 1 second
	unsigned short slowerTicks = 0, verySlowTicks = 0;
	unsigned short vacBuffer = 0;	

	while (1)
	{
		if (ticks > 245) // Ticks happen at 31372Hz so 245 is every 1/128 of a second
		{
			ticks -= 245;
			slowerTicks++;

			vacBuffer += ReadADC(8);

			// Prox pin tells us what state we're in. ADC around 900 for no EVSE connected, 600 when connected, 300 when connected and release switch pressed
			int prox;
			if (PCB_VERSION >= 11)
				prox = ReadADC(5);
			else
				prox = ReadADC(8);

			if (prox < 400 && state == IDLE) // Then cable has just been plugged in
			{
				state = CHARGING;
				PORTB |= (1<<PB1); // Charge Enable signal to EVMS
				PORTD |= (1<<PD5) | (1<<PD7); // PD5 = Ready signal to EVSE, PD7 is pull-down on prox resistor
				evmsPresent = false; // Reset this each time we start charging
				wattsAvailable = 0; // Reset this too
				error = false;
			}
			else if (prox > 400 && state == CHARGING) // Then cable has just been unplugged (and/or unlock button pressed)
			{
				state = IDLE;
				PORTB &= ~(1<<PB1); // Turn off Charge Enable signal to EVMS
				PORTD &= ~(1<<PD5); // Turn off Ready signal to EVSE
				PORTD &= ~(1<<PD7); // Turn off prox pull-down to reduce quiescent current consumption
			}

			if (prox < 100) // Something wrong
			{
				error = true;
				PORTD &= ~(1<<PD5); // Disable ready signal
			}

			smoothPilotTimer = (smoothPilotTimer*99L+100L*pilotTimer)/100L; // 100x the resolution and low pass filtering
			// What's the biggest value this could be? About 5000 * 100 * 100 = 50M, no problem for 32 bit variable
		}
		
		if (slowerTicks > 16) // 8hz ticks		
		{
			slowerTicks -= 16;
			verySlowTicks++;

			// Set LED - steady when full, flashing while charging
			if (!evmsPresent || evmsCommsTimeout > 0) PORTD &= ~(1<<PD4);

			if (state == CHARGING)
			{
				if (error)
				{
					if (verySlowTicks&0x01) PORTD |= (1<<PD4); // 4hz flashing if error
				}
				else if (verySlowTicks&0x04) PORTD |= (1<<PD4); // 1hz flashing if charging OK
			
				if (verySlowTicks&0x01) // Only transmit to EVMS at 4Hz, i.e every second time through loop
				{
					unsigned short supplyVoltage = vacBuffer/111; // With oversampling it's about 9.28 mV per ADC, i.e 1/108 V / ADC
						// Calibrated slightly different 10 1/108 due to internal resistance of AVR affecting voltage divider
					vacBuffer = 0; // reset					

					if (supplyVoltage < 10 || PCB_VERSION < 12) supplyVoltage = AC_VOLTAGE; // Default if no ADC available/connected
						
					if (pilotTimer < 660*8) // Only accept within certain range, this is roughly up to 40A
					{
						// EVSE signals how much current is available by 0.06A per 1uS, and we have 8 counts per uS
						long milliamps = smoothPilotTimer*3L/40L;
						wattsAvailable = (milliamps*(long)supplyVoltage/1000L);
						// Biggest value it could be during calc? 500K * 240 = 120M, no problem for 32-bit
					}
					else
					{
						error = true;
						PORTD &= ~(1<<PD5); // Disable ready signal
					}

					data[0] = wattsAvailable>>8; // High byte
					data[1] = wattsAvailable; // Low byte
					CanTX(45, data, 8);

					if (SEND_DIAGNOSTIC_VAC) // Pretend to be a current sensor, just for viewing AC voltage measured
					{
						long scaledValue = 8388608L + supplyVoltage*1000L;
						data[0] = scaledValue>>16;
						data[1] = scaledValue>>8;
						data[2] = scaledValue;
						CanTX(40, data, 8);
					}
				}
			}

			if (state == IDLE)
			{
				if (sleepTimer > 0)
					sleepTimer--;
				else
				{
					// Put micro to sleep to minimise quiescent current consumption, until negative transition on PROX pin
					PORTB |= (1<<PB0); // Turn off CAN transceiver too
					set_sleep_mode(SLEEP_MODE_PWR_DOWN); // Power down leaves only external INT0-3 and watchdog active
					sleep_enable();
					sleep_cpu();
					// Code execution stops here until falling edge on PROX pin
					sleep_disable(); // Wake from sleep
					PORTB &= ~(1<<PB0); // Turn on CAN transceiver	
					sleepTimer = 8;
				}
			}

			if (evmsPresent && evmsCommsTimeout > 0)
			{
				evmsCommsTimeout--;
				if (evmsCommsTimeout == 0) PORTD &= ~(1<<PD5); // Disable Ready signal if EVMS has turned off, means charge has finished
			}				
		}
	}

	return 0; // Never gets here but compiler wants to see it
}

void CanTX(long packetID, unsigned char* data, unsigned char length)
{
	CANPAGE = 0x00; // Select MOB0 for transmission
	while (CANEN2 & (1<<ENMOB0)); // Wait for MOB0 to be free
	CANSTMOB = 0x00;

	if (USE_29BIT_IDS) // CAN 2.0b is 29-bit IDs, CANIDT4 has bits 0-4 in top 5 bits, CANID3 has 5-12
	{
		CANIDT1 = packetID>>21;
		CANIDT2 = packetID>>13;
		CANIDT3 = packetID>>5;
		CANIDT4 = (packetID & 0b00011111)<<3;
	}
	else // CAN 2.0a is 11-bit IDs, IDT1 has top 8 bits, IDT2 has bottom three bits BUT at top of byte!
	{
		CANIDT1 = (packetID>>3); // Packet ID
		CANIDT2 = (packetID & 0x07)<<5;
		CANIDT3 = 0x00;
		CANIDT4 = 0x00;
	}

	for (int8_t i=0; i<length; i++) // trying single byte tx
		CANMSG = data[i];

	CANCDMOB = (1<<CONMOB0) | (length<<DLC0) | ((1<<IDE)*USE_29BIT_IDS); // Enable transmission, 8-bit data

	char timeout = 0;
	while (!(CANSTMOB & (1<<TXOK)) && !(CANSTMOB & (1<<AERR))) // Wait for transmission to finish (via setting of TXOK flag)
	{
		_delay_us(100);
		if (timeout++ > 100) break;
	}
	
	CANCDMOB = 0x00; // Disable transmission
	CANSTMOB = 0x00; // Clear TXOK flag
}

void SetupPorts()
{
	DDRB = 0b00000011; // PB1 = Charge Enable (signal to EVMS), PB0 pulls enable pin on CAN transceiver down (to turn on)
	DDRD = 0b10110000; // PD5 = ready signal (adjusts pilot impedance), PD4 = LED, PD7 is prox resistor pull-down
	
	// Timer0 is used for main loop ticks
	TCCR0B = (1<<CS00); // Clock on, 1:1 prescaler
	TIMSK0 = (1<<TOIE0); // Interrupt on overflow (increments ticks)

	// Timer1 is used for timing the control pilot pulse width	
	TCCR1B = (1<<CS10); // Enable timer 1 with no prescaler (=8Mhz)
	TIMSK1 = (1<<TOIE1); // Interrupt on overflow

	// INT0 is used to start/stop measurement of control pilot pulse
	EICRA = 0b00001001; // 01 for INT0 bits means interrupt on any logical change, 01 for INT1 used for wake on PROX falling edge
	EIMSK = 0b00000011; // Enables INT0 and INT1

	// ADC, used for measuring prox level (i.e plug detect)
	ADMUX |= ADC_VREF_TYPE;
	ADCSRA |= 0b10000110; // ADEN plus 110 prescaler = /64 for 125kHz ADC clock
	ADCSRB |= (1<<AREFEN);
	
	// CAN init stuff
	CANGCON = (1<<SWRES); // Software reset
	CANTCON = 0; // CAN timing prescaler set to 0
	if (CAN_BAUD_RATE == 1000)
	{
		CANBT1 = 0x00;
		CANBT2 = 0x04;
		CANBT3 = 0x12;
	}
	else
	{
		if (CAN_BAUD_RATE == 500)
			CANBT1 = 0x02;
		else if (CAN_BAUD_RATE == 250)
			CANBT1 = 0x06;
		else
			CANBT1 = 0x0E;
		CANBT2 = 0x04;
		CANBT3 = 0x13; // These set baud rate to 125kb/s (see ATmega16M1 manual, page 203)
	}

	for (int8_t mob=0; mob<6; mob++)
	{
		CANPAGE = (mob<<4); // Select MOB 0-5
		CANCDMOB = 0x00; // Disable MOB
		CANSTMOB = 0x00; // Clear MOB status register
	}

	CANPAGE = (1<<MOBNB0); // Select MOB1
	CANIE2 = (1<<IEMOB1); // Enable interrupts on MOB1 for reception and transmission
	CANGIE = (1<<ENIT) | (1<<ENRX); // Enable interrupts on receive and transmit
	CANIDM1 = CANIDM2 = CANIDM3 = CANIDM4 = 0x00; // CAN ID mask, zero will let all IDs pass
	CANCDMOB = (1<<CONMOB1) | (8<<DLC0) + ((1<<IDE)*USE_29BIT_IDS); // Enable reception, 11-bit IDE, 1-bit data length
	CANGCON |= (1<<ENASTB); // Enable mode. CAN channel enters enable mode after 11 recessive bits have been read
}
